/** Processing of complex constraints
 *
 * @author Ondrej Cekan
 * @date 20/03/2014
 * @update 05/07/2015
 *
     Copyright (C) 2015 Ondrej Cekan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <string>
#include <vector>


/** Processing of start() or end() constraint
 *
 * @param cons_name - string of the constraint
 * @param cons_string - syntax of string
 * @param nextgensyn - buffer of instructions
 * @return number of added instructions
 *
 */
unsigned cons_startend(string cons_name, Tsyn *cons_string, vector<pair<Tsyn *,string> > *nextgensyn)
{
    unsigned ninst = 0;

    for (gconsItr = gcons.rbegin(); gconsItr < gcons.rend(); gconsItr++)
    {
        if (gconsItr->op.compare(cons_name) == 0)
        {
            ninst++;
            switch (gconsItr->type)
            {
                case CSTR:
                {
                    cons_string->val.push_back(make_pair(gconsItr->rfrom,""));
                    nextgensyn->push_back(make_pair(&(*cons_string),""));
                } break;
                case CSYN:
                {
                    nextgensyn->push_back(make_pair(gconsItr->syn,""));
                } break;
                default: break;
            }
        }
    }

    return ninst;
}


/** Processing of probability pb() constraint
 *
 * @param syntax - instruction
 * @return success or failure
 *
 */
bool cons_probability(Tsyn *syntax, bool ignore_prob)
{
    for (syntax->consIt = syntax->cons.begin(); syntax->consIt < syntax->cons.end(); syntax->consIt++)
    { // constraints
        if (syntax->consIt->op.compare("pb") == 0)
        {
            int val = syntax->consIt->val;

            if (ignore_prob && val >= 0) return true;
            else if (!ignore_prob && val < 0) val = 0;

            if (abs(val) < (rand()%100 + 1)) return false;
        }
    }

    return true;
}


/** Substitution
 *
 * @param subs - substitutes
 * @param rate - vector of assigned substitutes
 * @param synt - instruction syntax
 * @param flag - flag
 * @param cons - constraint for the instruction
 * @param i - number of iteration of generation
 * @return success or failure
 *
 */
bool cons_substitute(vector<Tsubs> *subs, vector<pair<string, unsigned> > *rate, string *synt, string flag, vector<Tconsit> *cons, unsigned i, ofstream *fout)
{
    vector<Tsubs>::iterator subsIt;
    vector<pair<string, unsigned> >::iterator rateIt;
    vector<pair<string, string> > repldel;
    unsigned nrepl = 0;

    vector<pair<string, string> > wasrepl;


    for (subsIt = (*subs).begin(); subsIt < (*subs).end(); subsIt++)
    { // browse all possible substitutions
        unsigned watchdog = 0;
        bool replerr = false;

        while ((*synt).find(subsIt->name) != string::npos)
        { // substitution string was found
            watchdog++;

            // select random replacement
            string replacement = subsIt->subs[rand()%subsIt->subs.size()];
            bool is_rate = false;
            bool replwas = false;

            vector<pair<string, string> >::reverse_iterator replItr;


            // select a specific replacement
            if (!replerr && flag.compare("random") != 0)
            {
                for (replItr = repl.rbegin(); replItr < repl.rend(); replItr++)
                {
                    if (replItr->first == subsIt->name)
                    {
                        replacement = replItr->second;
                        replwas = true;

                        break;
                    }
                }
            }
            else if (flag.compare("random") == 0)
            { // select a random replacement
                bool valid = true;

                for (vector<pair<string, string> >::reverse_iterator rItr = repl.rbegin(); rItr < repl.rend(); rItr++)
                {
                    if (replacement == rItr->second)
                    {
                        valid = false;
                        break;
                    }
                }

                if (!valid) continue;
            }

            if (subsIt->is_input == true)
            { // is input, so no this value cannot be uninitialized - must be rated
                for (rateIt = (*rate).begin(); rateIt < (*rate).end(); rateIt++)
                { // browse rated variables
                    if ((*rateIt).first.compare(replacement) == 0)
                    { // check if the variable is rated, you could replace
                        if (i > (*rateIt).second)
                        {
                            is_rate = true;
                        }
                        break;
                    }
                }

                if (!is_rate)
                {
                    if ((*rate).size() <= 0 || watchdog > 500)
                    { // no valid substitution - take new instruction
                        while (nrepl > 0)
                        {
                            repl.erase(repl.begin());
                            nrepl--;
                        }

                        while (repldel.size() > 0)
                        {
                            repl.push_back(repldel.back());
                            repldel.pop_back();
                        }

                        return false;
                    }
                    else
                    { // take another valid substitution
                        replerr = true;
                        continue;
                    }
                }
            }
            else if (subsIt->is_input == false)
            { // is output, can be replaced without checking
                bool is_val = true;

                for (rateIt = (*rate).begin(); rateIt < (*rate).end(); rateIt++)
                { // check if it is in array
                    if ((*rateIt).first.compare(replacement) == 0)
                    {
                        is_rate = true;


                        if (i <= (*rateIt).second)
                        {
                            is_val = false;
                        }
                        break;
                    }
                }

                if (!is_val)
                {
                    replerr = true;
                    if (watchdog > 500) return false;
                    continue;
                }

                if (!is_rate)
                { // if it is not in array, add
                    rateIt = (*rate).begin();
                    (*rate).insert(rateIt, make_pair(replacement,i));
                }
            }


            bool tryagain = false;
            vector<Tconsit>::iterator consIt;
            for (consIt = (*cons).begin(); consIt < (*cons).end(); consIt++)
            {
                if (consIt->op.compare("nouse") == 0)
                { // nouse constraint - storing value
                    if (subsIt->name == consIt->rfrom)
                    {
                        for (rateIt = (*rate).begin(); rateIt < (*rate).end(); rateIt++)
                        {
                            if ((*rateIt).first.compare(replacement) == 0)
                            {
                                (*rateIt).second = i+consIt->val;
                                break;
                            }
                        }
                    }
                }

                if (consIt->op.compare("different") == 0)
                { // different constraint
                    vector<pair<string, string> >::iterator wasreplIt;
                    for (wasreplIt = wasrepl.begin(); wasreplIt < wasrepl.end(); wasreplIt++)
                    {
                        if (consIt->rfrom.compare(wasreplIt->first) == 0 || consIt->rto.compare(wasreplIt->first) == 0)
                        {
                            if (consIt->rto.compare(subsIt->name) == 0 || consIt->rfrom.compare(subsIt->name) == 0)
                            {
                                if (replacement.compare(wasreplIt->second) == 0)
                                {
                                    tryagain = true;
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            if (tryagain) continue;

            if (replwas)
            { // treatment
                repldel.insert(repldel.begin(), *replItr);
                repl.erase(--replItr.base());
            }


            wasrepl.insert(wasrepl.begin(),make_pair(subsIt->name,replacement));

            // replace
            *synt = replace_str(*synt, subsIt->name, replacement);
        }
    }

    return true;
}


/** Get random instruction
 *
 * @param synIt - instruction iterator
 * @param num - position of instruction
 * @return pointer to instruction
 *
 */
Tsyn *get_random_gen(map<string,Tsyn>::iterator synIt, unsigned num)
{
    unsigned synNumber = rand() % num;
    for (unsigned z = 0; z < synNumber; z++)
        synIt++;

    return &(synIt->second);
}


/** Create beforeinsert() structure
 *
 * @param syntax - instruction
 * @return vector of instructions
 *
 */
vector<pair<Tsyn *,string> > cons_beforeinsert(Tsyn *syntax)
{
    vector<pair<Tsyn *,string> > gback, gback2;

    for (syntax->consItr = syntax->cons.rbegin(); syntax->consItr < syntax->cons.rend(); syntax->consItr++)
    {
        if (syntax->consItr->op.compare("beforeinsert") == 0)
        {
            gback.push_back(make_pair(syntax->consItr->syn,""));
        }
    }

    gback.push_back(make_pair(syntax,""));

    if (gback2.size() > 0)
    {
        for (vector<pair<Tsyn *,string> >::iterator itemsIt = gback2.begin(); itemsIt < gback2.end(); itemsIt++)
            gback.push_back(*itemsIt);
    }

    return gback;
}

