/** Program generation for processors
 * - output is assembler code
 *
 * @author Ondrej Cekan
 * @date 20/02/2014
 * @update 05/07/2015
 *
     Copyright (C) 2015 Ondrej Cekan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <iostream>
#include <fstream>
#include <string>
#include <algorithm>
#include <cctype>
#include <vector>
#include <map>
#include <climits>
#include <sstream>
#include <time.h>

#include "main.h"
#include "error.cpp"
#include "global.cpp"
#include "parser.cpp"
#include "constraints.cpp"
#include "solver.cpp"


using namespace std;


/** Generator class
 */
class Generator
{
    fstream conf;           // problem description file
    fstream cons;           // constraints file
    string output_file;     // name of the output

    Solver *solv = NULL;    // solver = performs generation

    unsigned mlines;        // max lines

private:
        /** Load name of file from input and open it
         *
         * @param file - pointer to file
         * @param name - file name
         * @return boolean
         *
         */
        bool input(fstream *file, string name)
        {
            if (name.compare("conf") == 0)
            { // conf file
                cout << "Insert description file name: ";
                cin >> name;
                name.append(".conf");
            }
            else if (name.compare("cons") == 0)
            { // cons file
                cout << "Insert constraints file name: ";
                cin >> name;
                name.append(".cons");
            }

            (*file).open(name.c_str(), fstream::in);

            if (!(*file).is_open())
            { // cannot open file
                cerr << "Cannot open input file " << name << "." << endl;
                return false;
            }

            return true;
        }


public:
        /** Constructor
         *
         * @param conf_name - name of problem description file
         * @param cons_name - name of constraints file
         *
         */
        Generator (string conf_name, string cons_name)
        {
            if (!input(&conf,conf_name) ||
                !input(&cons,cons_name)) exit(1);

            // random generation initializing
            srand (time(NULL));
            rand();

            // parse file
            parser_conf(&conf);
            parser_cons(&cons);
        }

        /** Destructor
         */
        ~Generator()
        {
            // close file
            if (conf.is_open()) conf.close();
            if (cons.is_open()) cons.close();
        }

        /** Set output file name
         */
        void output(string name)
        {
            output_file = name;
        }

        /** Set max lines of output file
         */
        void maxlines(unsigned num)
        {
            mlines = num;
        }

        /** Generate one instruction
         */
        void generate()
        {
            if (solv == NULL) solv = new Solver (output_file,mlines);
            solv->generate();
        }

        /** Generate all instructions
         */
        void generate_all()
        {
            if (solv == NULL) solv = new Solver (output_file,mlines);
            solv->generate_all();
        }
};



/** Main program
 *
 * @param parameters of command line
 * @return boolean
 *
 */
int main(int argc, char *argv[])
{
    Generator gen ("risc.conf","risc.cons");
    gen.output("output.txt");

    if (argc > 1)
    {
        gen.output(argv[1]);
    }

    if (argc > 2)
    {
        IMAX = atoi(argv[2]);
    }

    gen.maxlines(IMAX);
    gen.generate();
    gen.generate();
    gen.generate_all();

    return 0;
}
