/** Parser for generator
 *
 * @author Ondrej Cekan
 * @date 20/03/2014
 * @update 05/07/2015
 *
     Copyright (C) 2015 Ondrej Cekan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <math.h>

/** Get Lex from string
 *
 * @param line - line from file
 * @param col - position in line
 * @return Lex
 *
 */
string lex(string *line, unsigned *col)
{
    string word;
    bool isstr = false;
    char prec, c;

    unsigned slen = (*line).length();
    for (; *col < slen; (*col)++)
    { // browsing the line
        c = (*line).at(*col);


        if (c == '"' || isstr)
        { // load string between quotation marks
            if (c != '"' || (c == '"' && prec == '\\'))
            {
                if (c == '"' && prec == '\\') word.at(word.size()-1) = c;
                else word.append(1,c);
            }

            if (c == '"' && prec != '\\' && isstr)
            {
                if (word.size() != 0)
                {
                    (*col)++;
                    return replace_str_all(word,"\\\\","\n"); // new line
                }
                else isstr = false;
            }
            else isstr = true;
        }
        else if (c == '/')
        {
            char nc = (*line).at((*col)+1);
            if (nc == '/')
            { // comments - leave whole line
                *col = slen;
                if (!word.empty()) return word;
                else               break;
            }
            if (!word.empty()) { return word; }
            else     { (*col)++; return string(1,c); }
        }
        else if (isalnum(c) || c == '-' || c == '_' || c == '+')
        { // valid chars
            word.append(1,c);
        }
        else if (isspace(c))
        { // spaces
            if (!word.empty()) { (*col)++; return word; }
            else { continue; }
        }
        else
        { // any other char
            if (!word.empty()) return word;
            else { (*col)++; return string(1,c); }
        }
        prec = c;
    }

    return "\n";
}


/** Parser for configuration file
 *
 * @param file - pointer to input file
 *
 */
void parser_conf(fstream *file)
{
    string line;                // current line
    string word;                // current Lex
    unsigned row = 0;           // number of line
    pstate state = INIT;        // actual state

    while(getline(*file,line))
    { // get line from file
        line.append("\n");      // add new line for checking end of line

        var_type vtype = VINT;
        unsigned col = 0;       // position in line
        unsigned nsubs = 0;     // number of substitute words in line
        vector<string> nsyn;    // number of syntax names in line
        row++;

        vector<string> lsubs;   // substitute names for substitution
        bool wordrd = false;    // flag reading word

        while ((word.assign(lex(&line,&col)).compare("\n")) != 0)
        { // get Lex from line
            switch (state)
            {
            /**< INIT */
                case INIT:
                {
                         if (word.compare("substitute") == 0)  state = SUBS;
                    else if (word.compare("variable") == 0)    state = VAR;
                    else if (word.compare("syntax") == 0)      state = SYN;
                    else error(UKEYWORD,row,col,word);

                    word.assign(lex(&line,&col));
                    if (word.compare("{") != 0) error(UWORD,row,col,word);
                } break;

            /**< SUBSTITUTE */
                case SUBS:
                {
                    wordrd = false;

                    if (word.compare(",") == 0)      continue;
                    else if (word.compare("{") == 0) state = SUBSSUBS;
                    else if (word.compare("}") == 0) state = INIT;
                    else if (is_keyword(word))
                    {
                        Tsubs item;
                        item.is_input = true;
                        item.name.assign(word);

                        subsIt = subs.begin();
                        subs.insert(subsIt, item);
                        nsubs++;
                    }
                    else error(UWORD,row,col,word);
                } break;

                case SUBSSUBS:
                {
                    if (word.compare("|") == 0 || word.compare("}") == 0)
                    {
                        if (word.compare("}") == 0)
                        {
                            if (nsubs > 0)
                            {
                                unsigned i = 0;
                                for (subsIt = subs.begin(); i < nsubs && subsIt < subs.end(); subsIt++)
                                {
                                    subsIt->subs = lsubs;
                                    i++;
                                }
                            }
                            state = SUBS;
                        }
                        wordrd = false;
                    }
                    else if (!wordrd)
                    {
                        vector<string>::iterator lsubsIt = lsubs.begin();
                        lsubs.insert(lsubsIt,word);
                        wordrd = true;
                    }
                    else error(UWORD,row,col,word);
                } break;

            /**< VARIABLE */
                case VAR:
                case VARNUM:
                {
                         if (word.compare("VAR8") == 0)  vtype = VINT;
                    else if (word.compare("VAR16") == 0) vtype = VINT16;
                    else if (word.compare("VAR14") == 0) vtype = VINT14;
                    else if (word.compare("BOOL") == 0)  vtype = VBOOL;
                    else if (word.compare("STR") == 0)   vtype = VSTR;
                    else if (word.compare("}") == 0)
                    {
                        state = INIT;
                        break;
                    }
                    else if (word.compare("{") == 0 && state == VARNUM)
                    {
                        state = VARMIN;
                        continue;
                    }
                    else error(UWORD,row,col,word);

                    state = VARNAME;
                } break;

                case VARNAME:
                { // name of variable
                    if (is_keyword(word))
                    {
                        Tvar item;
                        item.type = vtype;
                        item.name.assign(word);

                        varIt = var.begin();
                        varIt = var.insert(varIt, item);

                        state = VARNUM;
                    }
                    else error(UWORD,row,col,word);
                } break;

                case VARMIN:
                { // min value for variable
                    if (is_numeric(word) || word.compare("cpos") == 0)
                    {
                        Tconsit item;
                        item.op = "min";
                        if (word.compare("cpos") == 0)
                        {
                            item.str = "cpos";
                        }
                        else
                        {
                            item.val = atoi(word.c_str());
                        }

                        varIt->cons.insert(varIt->cons.begin(),item);

                        word.assign(lex(&line,&col));
                        if (word.compare(".") == 0)
                        {
                            word.assign(lex(&line,&col));
                            if (word.compare(".") == 0)
                            {
                                state = VARMAX;
                                continue;
                            }
                        }

                        error(UWORD,row,col,word);
                    }
                } break;

                case VARMAX:
                { // max value for variable
                    if (is_numeric(word) || word.compare("cpos") == 0)
                    {
                        Tconsit item;
                        item.op = "max";
                        if (word.compare("cpos") == 0)
                        {
                            item.str = "cpos";
                        }
                        else
                        {
                            item.val = atoi(word.c_str());
                        }

                        varIt->cons.insert(varIt->cons.begin(),item);

                        word.assign(lex(&line,&col));
                        if (word.compare("}") == 0)
                        {
                            state = VAR;
                        }
                        else error(UWORD,row,col,word);
                    }
                } break;

            /**< SYNTAX */
                case SYN:
                {
                    wordrd = false;

                    if (word.compare(",") == 0)      continue;
                    else if (word.compare("{") == 0) state = SYNSYNTAX;
                    else if (word.compare("}") == 0) state = INIT;
                    else if (is_keyword(word))
                    {
                        Tsyn item;
                        item.name.assign(word);

                        synIt = syn.begin();
                        syn.insert(synIt, pair<string,Tsyn>(word, item));
                        nsyn.insert(nsyn.begin(), word);
                    }
                    else error(UWORD,row,col,word);
                } break;

                case SYNSYNTAX:
                {
                    if (word.compare("}") == 0) state = SYN;
                    else if (!wordrd)
                    {
                        if (nsyn.size() > 0)
                        {
                            for (vector<string>::iterator nsynIt = nsyn.begin(); nsynIt < nsyn.end(); nsynIt++)
                            {
                                synIt = syn.find(*nsynIt);
                                synIt->second.syntax.assign(replace_str(word,"this",synIt->first));
                            }
                            nsyn.clear();
                        }
                        wordrd = true;
                    }
                    else error(UWORD,row,col,word);
                } break;

            /**< DEFAULT */
                default: error(UKNOWN,row,col,word); break;
            }
        }

    }
}


/** Parser for constraint file
 *
 * @param file - pointer to input file
 *
 */
void parser_cons(fstream *file)
{
    string line;                 // current line
    string word;                 // current Lex
    unsigned row = 0;           // number of line
    pstate state = INIT;        // actual state

    while(getline(*file,line))
    { // get line from file
        line.append("\n");      // add new line for checking end of line

        Tvar *ivar = NULL;      // pointer for ivar data type
        Tsyn *isyn = NULL;      // pointer for isyn data type
        string op;              // actual constraint name
        unsigned col = 0;       // position in line
        row++;

        while ((word.assign(lex(&line,&col)).compare("\n")) != 0)
        { // get Lex from line
            switch (state)
            {
            /**< INIT */
                case INIT:
                {
                      if (word.compare("constraint") == 0)  state = CONS;
                    else error(UWORD,row,col,word);

                    word.assign(lex(&line,&col));
                    if (word.compare("{") != 0) error(UWORD,row,col,word);
                } break;

            /**< CONSTRAINT SOLVER */
                case CONS:
                {
                    if (is_constraint(word))
                    {
                        op.assign(word);

                        state = CONS1;
                        if (word.compare("output") == 0 ||
                            word.compare("nocare") == 0 ||
                            word.compare("outall") == 0 ||
                            word.compare("start") == 0 ||
                            word.compare("nlines") == 0 ||
                            word.compare("minval") == 0 ||
                            word.compare("maxval") == 0 ||
                            word.compare("cposbalance") == 0 ||
                            word.compare("linedelimiter") == 0 ||
                            word.compare("useonly") == 0 ||
                            word.compare("nslots") == 0 ||
                            word.compare("end") == 0) state = CONSSPECIAL;

                        word.assign(lex(&line,&col));
                        if (word.compare("(") != 0) error(UWORD,row,col,word);
                    }
                    else if (word.compare("}") == 0) state = INIT;
                    else error(UWORD,row,col,word);
                } break;

                case CONSSPECIAL:
                { // global constraint or constraint with only one parameter
                    if (op.compare("output") == 0 || op.compare("nocare") == 0)
                    {
                        if (is_keyword(word))
                        {
                            for (subsIt = subs.begin(); subsIt < subs.end(); subsIt++)
                            {
                                if (subsIt->name.compare(word) == 0)
                                {
                                    if (op.compare("nocare") == 0) subsIt->is_input = 2; // 2 = nocare
                                    else subsIt->is_input = false;
                                    break;
                                }
                            }
                        }
                        else error(UWORD,row,col,word);
                    }
                    else if (op.compare("nlines") == 0)
                    {
                        MINROWS = atoi(word.c_str());

                        word.assign(lex(&line,&col));
                        word.assign(lex(&line,&col));

                        MAXROWS = atoi(word.c_str());

                        IMAX = (rand() % (MAXROWS+1-MINROWS)) + MINROWS;
                    }
                    else if (op.compare("nslots") == 0)
                    {
                        NSLOTS = atoi(word.c_str());
                    }
                    else if (op.compare("linedelimiter") == 0)
                    {
                        DELIMITER = word;
                    }
                    else if (op.compare("minval") == 0 || op.compare("maxval") == 0)
                    {
                        if (is_keyword(word))
                        {
                            for (varIt = var.begin(); varIt < var.end(); varIt++)
                            {
                                if (varIt->name.compare(word) == 0)
                                {
                                    word.assign(lex(&line,&col));
                                    word.assign(lex(&line,&col));

                                    if (is_numeric(word))
                                    {
                                        Tconsit item;
                                        if (op.compare("minval") == 0) item.op = "min";
                                        else                           item.op = "max";
                                        item.val = atoi(word.c_str());

                                        varIt->cons.insert(varIt->cons.begin(),item);
                                    }
                                    else if (word.compare("nlines") == 0 || word.compare("nlinesrest") == 0)
                                    {
                                        Tconsit item;
                                        if (op.compare("minval") == 0) item.op = "min";
                                        else                           item.op = "max";
                                        item.val = IMAX;


                                        if (word.compare("nlinesrest") == 0)
                                        {
                                            item.str = "rest";
                                        }


                                        word.assign(lex(&line,&col));
                                        if (word.compare("(") == 0)
                                        {
                                            word.assign(lex(&line,&col));
                                            if (is_numeric(word))
                                            {
                                                item.val = ceil(IMAX/100*atoi(word.c_str()));
                                                word.assign(lex(&line,&col));
                                            }
                                        }

                                        varIt->cons.insert(varIt->cons.begin(),item);
                                    }
                                }
                            }
                        }
                        else error(UWORD,row,col,word);
                    }
                    else if (op.compare("cposbalance") == 0)
                    {
                        if (is_numeric(word))
                        {
                            CPBALANCE = atoi(word.c_str());
                        }
                    }
                    else if (op.compare("useonly") == 0)
                    {
                        Tconsit item;
                        item.op = op;
                        item.type = CNONE;

                        synIt = syn.find(word);
                        if (synIt != syn.end())
                        {
                            item.type = CSYN;
                            item.syn = &(synIt->second);
                        }

                        if (item.type == CNONE) error(USYN,row,col,word);

                        gcons.insert(gcons.begin(),item);
                    }
                    else
                    { // classic approach for constraint processing
                        Tconsit item;
                        item.op = op;
                        item.type = CNONE;

                        for (varIt = var.begin(); varIt < var.end(); varIt++)
                        { // variable
                            if (varIt->name.compare(word) == 0)
                            {
                                item.type = CVAR;
                                item.var = &(*varIt);
                                break;
                            }
                        }

                        if (item.type == CNONE)
                        { // syntaxe
                            synIt = syn.find(word);
                            if (synIt != syn.end())
                            {
                                item.type = CSYN;
                                item.syn = &(synIt->second);
                            }
                        }

                        if (item.type == CNONE)
                        { // string
                            item.type = CSTR;
                            item.rfrom = word;
                        }

                        gcons.insert(gcons.begin(),item);
                    }

                    state = CONS;

                    word.assign(lex(&line,&col));
                    if (word.compare(")") != 0) error(UWORD,row,col,word);
                } break;

            /**< CONSTRAINT */
                case CONS1:
                { // first parameter
                    Tconsit item;
                    item.op = op;

                    for (varIt = var.begin(); varIt < var.end(); varIt++)
                    { // variable
                        if (varIt->name.compare(word) == 0)
                        {
                            varIt->cons.insert(varIt->cons.begin(),item);
                            ivar = &(*varIt);
                            break;
                        }
                    }

                    if (ivar == NULL)
                    {
                        synIt = syn.find(word);
                        if (synIt != syn.end())
                        {
                            synIt->second.cons.insert(synIt->second.cons.begin(), item);
                            isyn = &(synIt->second);
                        }
                    }

                    word.assign(lex(&line,&col));

                    if (word.compare(",") == 0)      state = CONS2;
                    else if (word.compare("(") == 0) state = CONSFUNC1;
                    else if (word.compare(")") == 0) state = CONS;
                    else error(UWORD,row,col,word);
                } break;

                case CONS2:
                { // second parameter
                    bool is_assign = false;

                    for (varIt = var.begin(); varIt < var.end(); varIt++)
                    { // variable
                        if (varIt->name.compare(word) == 0)
                        {
                            if (ivar != NULL)
                            {
                                ivar->consIt = ivar->cons.begin();
                                ivar->consIt->var = &(*varIt);
                                ivar->consIt->type = CVAR;
                            }
                            else if (isyn != NULL)
                            {
                                isyn->consIt = isyn->cons.begin();
                                isyn->consIt->var = &(*varIt);
                                isyn->consIt->type = CVAR;
                            }
                            is_assign = true;
                            break;
                        }
                    }

                    if (!is_assign && is_numeric(word))
                    { // number
                        if (ivar != NULL)
                        {
                            ivar->consIt = ivar->cons.begin();
                            ivar->consIt->val = atoi(word.c_str());
                            ivar->consIt->type = CNUM;
                        }
                        else if (isyn != NULL)
                        {
                            isyn->consIt = isyn->cons.begin();
                            isyn->consIt->val = atoi(word.c_str());
                            isyn->consIt->type = CNUM;
                        }
                        is_assign = true;
                    }

                    if (!is_assign)
                    { // syntaxe
                        synIt = syn.find(word);
                        if (synIt != syn.end())
                        {
                            if (ivar != NULL)
                            {
                                ivar->consIt = ivar->cons.begin();
                                ivar->consIt->syn = &(synIt->second);
                                ivar->consIt->type = CSYN;
                            }
                            else if (isyn != NULL)
                            {
                                isyn->consIt = isyn->cons.begin();
                                isyn->consIt->syn = &(synIt->second);
                                isyn->consIt->type = CSYN;
                            }
                            is_assign = true;
                        }
                    }

                    if (!is_assign)
                    { // string
                        if (ivar != NULL)
                        {
                            ivar->consIt = ivar->cons.begin();
                            ivar->consIt->type = CSTR;
                            ivar->consIt->str.assign(word);
                        }
                        else if (isyn != NULL)
                        {
                            isyn->consIt = isyn->cons.begin();
                            isyn->consIt->type = CSTR;
                            isyn->consIt->str.assign(word);
                        }
                    }

                    word.assign(lex(&line,&col));

                    if (word.compare(")") == 0)      state = CONS;
                    else if (word.compare("(") == 0) state = CONSFUNC2;
                    else error(UWORD,row,col,word);
                } break;

                case CONSFUNC1:
                { // parameter of the first parameter
                    if (is_keyword(word) && isyn != NULL)
                    {
                        isyn->consIt = isyn->cons.begin();

                        bool isvar = false;
                        if (op.compare("storeval") == 0)
                        {
                            for (varIt = var.begin(); varIt < var.end(); varIt++)
                            { // variable
                                if (varIt->name.compare(word) == 0)
                                {
                                    isvar = true;
                                    isyn->consIt->var = &(*varIt);
                                    isyn->consIt->val = 1;

                                    word.assign(lex(&line,&col));
                                    if (word.compare("(") == 0)
                                    {
                                        word.assign(lex(&line,&col));
                                        if (is_numeric(word))
                                        {
                                            isyn->consIt->val = atoi(word.c_str());
                                            word.assign(lex(&line,&col));
                                        }
                                    }

                                    break;
                                }
                            }
                        }

                        if (op.compare("storeval") != 0 || !isvar) isyn->consIt->rfrom.assign(word);
                    }
                    else error(UWORD,row,col,word);

                    word.assign(lex(&line,&col));

                    if (word.compare(",") == 0)
                    {
                        word.assign(lex(&line,&col));

                        if (is_keyword(word))
                        {
                            bool isvar = false;
                            if (op.compare("storeval") == 0)
                            {
                                for (varIt = var.begin(); varIt < var.end(); varIt++)
                                { // variable
                                    if (varIt->name.compare(word) == 0)
                                    {
                                        isvar = true;
                                        isyn->consIt->var2 = &(*varIt);
                                        isyn->consIt->val2 = 1;

                                        word.assign(lex(&line,&col));
                                        if (word.compare("(") == 0)
                                        {
                                            word.assign(lex(&line,&col));
                                            if (is_numeric(word))
                                            {
                                                isyn->consIt->val2 = atoi(word.c_str());
                                                word.assign(lex(&line,&col));
                                            }
                                        }

                                        break;
                                    }
                                }
                            }

                            if (op.compare("storeval") != 0 || !isvar) isyn->consIt->rto.assign(word);
                        }
                        else error(UWORD,row,col,word);

                        word.assign(lex(&line,&col));
                    }


                    if (word.compare(")") == 0)
                    {
                        word.assign(lex(&line,&col));
                        if (word.compare(",") == 0)
                        {
                            state = CONS2;
                            break;
                        }
                        else if (word.compare(")") == 0)
                        {
                            state = CONS;
                            break;
                        }
                    }
                    error(UWORD,row,col,word);
                } break;


                case CONSFUNC2:
                { // parameter of the second parameter
                    if (is_keyword(word) && isyn != NULL)
                    {
                        isyn->consIt = isyn->cons.begin();
                        isyn->consIt->rto.assign(word);
                    }
                    else error(UWORD,row,col,word);

                    word.assign(lex(&line,&col));
                    if (word.compare(")") == 0)
                    {
                        word.assign(lex(&line,&col));
                        if (word.compare(")") == 0)
                        {
                            state = CONS;
                            break;
                        }
                    }
                    error(UWORD,row,col,word);
                } break;

            /**< DEFAULT */
                default: error(UKNOWN,row,col,word); break;
            }
        }
    }
}
